@@ -20,10 +20,9 @@
 package services
 
 import (
-	"encoding/json"
+	"context"
 	"fmt"
 	"path/filepath"
-	"runtime"
 	"strconv"
 	"strings"
 
@@ -51,20 +50,23 @@ import (
 	virtconfig "kubevirt.io/kubevirt/pkg/virt-config"
 )
 
-const configMapName = "kubevirt-config"
 const KvmDevice = "devices.kubevirt.io/kvm"
 const TunDevice = "devices.kubevirt.io/tun"
 const VhostNetDevice = "devices.kubevirt.io/vhost-net"
 
 const debugLogs = "debugLogs"
+const logVerbosity = "logVerbosity"
 const virtiofsDebugLogs = "virtiofsdDebugLogs"
 
 const MultusNetworksAnnotation = "k8s.v1.cni.cncf.io/networks"
 
-const CAP_NET_ADMIN = "NET_ADMIN"
-const CAP_NET_RAW = "NET_RAW"
-const CAP_SYS_ADMIN = "SYS_ADMIN"
-const CAP_SYS_NICE = "SYS_NICE"
+const (
+	CAP_NET_ADMIN    = "NET_ADMIN"
+	CAP_NET_RAW      = "NET_RAW"
+	CAP_SYS_ADMIN    = "SYS_ADMIN"
+	CAP_SYS_NICE     = "SYS_NICE"
+	CAP_SYS_RESOURCE = "SYS_RESOURCE"
+)
 
 // LibvirtStartupDelay is added to custom liveness and readiness probes initial delay value.
 // Libvirt needs roughly 10 seconds to start.
@@ -72,9 +74,10 @@ const LibvirtStartupDelay = 10
 
 //These perfixes for node feature discovery, are used in a NodeSelector on the pod
 //to match a VirtualMachineInstance CPU model(Family) and/or features to nodes that support them.
-const NFD_CPU_MODEL_PREFIX = "feature.node.kubernetes.io/cpu-model-"
-const NFD_CPU_FEATURE_PREFIX = "feature.node.kubernetes.io/cpu-feature-"
-const NFD_KVM_INFO_PREFIX = "feature.node.kubernetes.io/kvm-info-cap-hyperv-"
+const NFD_CPU_MODEL_PREFIX = "cpu-model.node.kubevirt.io/"
+const NFD_CPU_FEATURE_PREFIX = "cpu-feature.node.kubevirt.io/"
+const NFD_KVM_INFO_PREFIX = "hyperv.node.kubevirt.io/"
+const IntelVendorName = "Intel"
 
 const MULTUS_RESOURCE_NAME_ANNOTATION = "k8s.v1.cni.cncf.io/resourceName"
 const MULTUS_DEFAULT_NETWORK_CNI_ANNOTATION = "v1.multus-cni.io/default-network"
@@ -84,12 +87,23 @@ const ISTIO_KUBEVIRT_ANNOTATION = "traffic.sidecar.istio.io/kubevirtInterfaces"
 
 const ENV_VAR_LIBVIRT_DEBUG_LOGS = "LIBVIRT_DEBUG_LOGS"
 const ENV_VAR_VIRTIOFSD_DEBUG_LOGS = "VIRTIOFSD_DEBUG_LOGS"
+const ENV_VAR_VIRT_LAUNCHER_LOG_VERBOSITY = "VIRT_LAUNCHER_LOG_VERBOSITY"
+
 const ENV_VAR_POD_NAME = "POD_NAME"
 
+// extensive log verbosity threshold after which libvirt debug logs will be enabled
+const EXT_LOG_VERBOSITY_THRESHOLD = 5
+
+const ephemeralStorageOverheadSize = "50M"
+
 type TemplateService interface {
 	RenderLaunchManifest(*v1.VirtualMachineInstance) (*k8sv1.Pod, error)
-	RenderHotplugAttachmentPodTemplate(volume *v1.Volume, ownerPod *k8sv1.Pod, vmi *v1.VirtualMachineInstance, pvcName string, isBlock bool) (*k8sv1.Pod, error)
+	RenderHotplugAttachmentPodTemplate(volume *v1.Volume, ownerPod *k8sv1.Pod, vmi *v1.VirtualMachineInstance, pvcName string, isBlock bool, tempPod bool) (*k8sv1.Pod, error)
 	RenderLaunchManifestNoVm(*v1.VirtualMachineInstance) (*k8sv1.Pod, error)
+	GetLauncherImage() string
+	GetCpuArch() string
+	IsPPC64() bool
+	IsARM64() bool
 }
 
 type templateService struct {
@@ -104,6 +118,7 @@ type templateService struct {
 	virtClient                 kubecli.KubevirtClient
 	clusterConfig              *virtconfig.ClusterConfig
 	launcherSubGid             int64
+	cpuArch                    string
 }
 
 type PvcNotFoundError error
@@ -134,42 +149,48 @@ func makeHVFeatureLabelTable(vmi *v1.VirtualMachineInstance) []hvFeatureLabel {
 	// to learn about dependencies between enlightenments
 
 	hyperv := vmi.Spec.Domain.Features.Hyperv // shortcut
+
+	syNICTimer := &v1.FeatureState{}
+	if hyperv.SyNICTimer != nil {
+		syNICTimer.Enabled = hyperv.SyNICTimer.Enabled
+	}
+
 	return []hvFeatureLabel{
-		hvFeatureLabel{
+		{
 			Feature: hyperv.VPIndex,
 			Label:   "vpindex",
 		},
-		hvFeatureLabel{
+		{
 			Feature: hyperv.Runtime,
 			Label:   "runtime",
 		},
-		hvFeatureLabel{
+		{
 			Feature: hyperv.Reset,
 			Label:   "reset",
 		},
-		hvFeatureLabel{
+		{
 			// TODO: SyNIC depends on vp-index on QEMU level. We should enforce this constraint.
 			Feature: hyperv.SyNIC,
 			Label:   "synic",
 		},
-		hvFeatureLabel{
+		{
 			// TODO: SyNICTimer depends on SyNIC and Relaxed. We should enforce this constraint.
-			Feature: hyperv.SyNICTimer,
+			Feature: syNICTimer,
 			Label:   "synictimer",
 		},
-		hvFeatureLabel{
+		{
 			Feature: hyperv.Frequencies,
 			Label:   "frequencies",
 		},
-		hvFeatureLabel{
+		{
 			Feature: hyperv.Reenlightenment,
 			Label:   "reenlightenment",
 		},
-		hvFeatureLabel{
+		{
 			Feature: hyperv.TLBFlush,
 			Label:   "tlbflush",
 		},
-		hvFeatureLabel{
+		{
 			Feature: hyperv.IPI,
 			Label:   "ipi",
 		},
@@ -188,6 +209,11 @@ func getHypervNodeSelectors(vmi *v1.VirtualMachineInstance) map[string]string {
 			nodeSelectors[NFD_KVM_INFO_PREFIX+hv.Label] = "true"
 		}
 	}
+
+	if vmi.Spec.Domain.Features.Hyperv.EVMCS != nil {
+		nodeSelectors[v1.CPUModelVendorLabel+IntelVendorName] = "true"
+	}
+
 	return nodeSelectors
 }
 
@@ -261,6 +287,28 @@ func SetNodeAffinityForForbiddenFeaturePolicy(vmi *v1.VirtualMachineInstance, po
 	}
 }
 
+func sysprepVolumeSource(sysprepVolume v1.SysprepSource) (k8sv1.VolumeSource, error) {
+	logger := log.DefaultLogger()
+	if sysprepVolume.Secret != nil {
+		return k8sv1.VolumeSource{
+			Secret: &k8sv1.SecretVolumeSource{
+				SecretName: sysprepVolume.Secret.Name,
+			},
+		}, nil
+	} else if sysprepVolume.ConfigMap != nil {
+		return k8sv1.VolumeSource{
+			ConfigMap: &k8sv1.ConfigMapVolumeSource{
+				LocalObjectReference: k8sv1.LocalObjectReference{
+					Name: sysprepVolume.ConfigMap.Name,
+				},
+			},
+		}, nil
+	}
+	errorStr := fmt.Sprintf("Sysprep must have Secret or ConfigMap reference set %v", sysprepVolume)
+	logger.Errorf(errorStr)
+	return k8sv1.VolumeSource{}, fmt.Errorf(errorStr)
+}
+
 // Request a resource by name. This function bumps the number of resources,
 // both its limits and requests attributes.
 //
@@ -308,6 +356,11 @@ func requestResource(resources *k8sv1.ResourceRequirements, resourceName string)
 		resources.Requests[name] = unitQuantity
 	}
 }
+
+func (t *templateService) GetLauncherImage() string {
+	return t.launcherImage
+}
+
 func (t *templateService) RenderLaunchManifestNoVm(vmi *v1.VirtualMachineInstance) (*k8sv1.Pod, error) {
 	return t.renderLaunchManifest(vmi, true)
 }
@@ -315,6 +368,28 @@ func (t *templateService) RenderLaunchManifest(vmi *v1.VirtualMachineInstance) (
 	return t.renderLaunchManifest(vmi, false)
 }
 
+func (t *templateService) GetCpuArch() string {
+	return t.getCpuArch()
+}
+
+func (t *templateService) getCpuArch() string {
+	return t.cpuArch
+}
+
+func (t *templateService) IsPPC64() bool {
+	if t.cpuArch == "ppc64le" {
+		return true
+	}
+	return false
+}
+
+func (t *templateService) IsARM64() bool {
+	if t.cpuArch == "arm64" {
+		return true
+	}
+	return false
+}
+
 func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, tempPod bool) (*k8sv1.Pod, error) {
 	precond.MustNotBeNil(vmi)
 	domain := precond.MustNotBeEmpty(vmi.GetObjectMeta().GetName())
@@ -329,7 +404,7 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 	var imagePullSecrets []k8sv1.LocalObjectReference
 
 	// Need to run in privileged mode in Power or libvirt will fail to lock memory for VMI
-	if runtime.GOARCH == "ppc64le" {
+	if t.IsPPC64() {
 		privileged = true
 	}
 
@@ -374,23 +449,6 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 		},
 	})
 
-	if util.IsVFIOVMI(vmi) && !util.IsSRIOVVmi(vmi) {
-		// libvirt needs this volume to access PCI device config;
-		// note that the volume should not be read-only because libvirt
-		// opens the config for writing
-		volumeMounts = append(volumeMounts, k8sv1.VolumeMount{
-			Name:      "pci-devices",
-			MountPath: "/sys/devices/",
-		})
-		volumes = append(volumes, k8sv1.Volume{
-			Name: "pci-devices",
-			VolumeSource: k8sv1.VolumeSource{
-				HostPath: &k8sv1.HostPathVolumeSource{
-					Path: "/sys/devices/",
-				},
-			},
-		})
-	}
 	serviceAccountName := ""
 
 	for _, volume := range vmi.Spec.Volumes {
@@ -568,6 +626,12 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 					SubPath:   "userdata",
 					ReadOnly:  true,
 				})
+				volumeMounts = append(volumeMounts, k8sv1.VolumeMount{
+					Name:      volumeName,
+					MountPath: filepath.Join(config.SecretSourceDir, volume.Name, "userData"),
+					SubPath:   "userData",
+					ReadOnly:  true,
+				})
 			}
 			if volume.CloudInitNoCloud.NetworkDataSecretRef != nil {
 				// attach a secret referenced by the networkdata
@@ -586,9 +650,33 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 					SubPath:   "networkdata",
 					ReadOnly:  true,
 				})
+				volumeMounts = append(volumeMounts, k8sv1.VolumeMount{
+					Name:      volumeName,
+					MountPath: filepath.Join(config.SecretSourceDir, volume.Name, "networkData"),
+					SubPath:   "networkData",
+					ReadOnly:  true,
+				})
 			}
 		}
 
+		if volume.Sysprep != nil {
+			var volumeSource k8sv1.VolumeSource
+			// attach a Secret or ConfigMap referenced by the user
+			volumeSource, err := sysprepVolumeSource(*volume.Sysprep)
+			if err != nil {
+				return nil, err
+			}
+			volumes = append(volumes, k8sv1.Volume{
+				Name:         volume.Name,
+				VolumeSource: volumeSource,
+			})
+			volumeMounts = append(volumeMounts, k8sv1.VolumeMount{
+				Name:      volume.Name,
+				MountPath: filepath.Join(config.SysprepSourceDir, volume.Name),
+				ReadOnly:  true,
+			})
+		}
+
 		if volume.CloudInitConfigDrive != nil {
 			if volume.CloudInitConfigDrive.UserDataSecretRef != nil {
 				// attach a secret referenced by the user
@@ -607,6 +695,12 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 					SubPath:   "userdata",
 					ReadOnly:  true,
 				})
+				volumeMounts = append(volumeMounts, k8sv1.VolumeMount{
+					Name:      volumeName,
+					MountPath: filepath.Join(config.SecretSourceDir, volume.Name, "userData"),
+					SubPath:   "userData",
+					ReadOnly:  true,
+				})
 			}
 			if volume.CloudInitConfigDrive.NetworkDataSecretRef != nil {
 				// attach a secret referenced by the networkdata
@@ -625,6 +719,12 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 					SubPath:   "networkdata",
 					ReadOnly:  true,
 				})
+				volumeMounts = append(volumeMounts, k8sv1.VolumeMount{
+					Name:      volumeName,
+					MountPath: filepath.Join(config.SecretSourceDir, volume.Name, "networkData"),
+					SubPath:   "networkData",
+					ReadOnly:  true,
+				})
 			}
 		}
 	}
@@ -670,7 +770,7 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 	gracePeriodKillAfter := gracePeriodSeconds + int64(15)
 
 	// Get memory overhead
-	memoryOverhead := getMemoryOverhead(vmi)
+	memoryOverhead := getMemoryOverhead(vmi, t.cpuArch)
 
 	// Consider CPU and memory requests and limits for pod scheduling
 	resources := k8sv1.ResourceRequirements{}
@@ -689,7 +789,7 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 		if vcpus != 0 && cpuAllocationRatio > 0 {
 			val := float64(vcpus) / float64(cpuAllocationRatio)
 			vcpusStr := fmt.Sprintf("%g", val)
-			if val < 0 {
+			if val < 1 {
 				val *= 1000
 				vcpusStr = fmt.Sprintf("%gm", val)
 			}
@@ -706,6 +806,18 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 		resources.Limits[key] = value
 	}
 
+	// Add ephemeral storage request to container to be used by Kubevirt. This amount of ephemeral storage
+	// should be added to the user's request.
+	ephemeralStorageOverhead := resource.MustParse(ephemeralStorageOverheadSize)
+	ephemeralStorageRequested := resources.Requests[k8sv1.ResourceEphemeralStorage]
+	ephemeralStorageRequested.Add(ephemeralStorageOverhead)
+	resources.Requests[k8sv1.ResourceEphemeralStorage] = ephemeralStorageRequested
+
+	if ephemeralStorageLimit, ephemeralStorageLimitDefined := resources.Limits[k8sv1.ResourceEphemeralStorage]; ephemeralStorageLimitDefined {
+		ephemeralStorageLimit.Add(ephemeralStorageOverhead)
+		resources.Limits[k8sv1.ResourceEphemeralStorage] = ephemeralStorageLimit
+	}
+
 	// Consider hugepages resource for pod scheduling
 	if vmi.Spec.Domain.Memory != nil && vmi.Spec.Domain.Memory.Hugepages != nil {
 		hugepageType := k8sv1.ResourceName(k8sv1.ResourceHugePagesPrefix + vmi.Spec.Domain.Memory.Hugepages.PageSize)
@@ -871,7 +983,7 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 	// Add ports from interfaces to the pod manifest
 	ports := getPortsFromVMI(vmi)
 
-	capabilities := getRequiredCapabilities(vmi)
+	capabilities := getRequiredCapabilities(vmi, t.clusterConfig)
 
 	networkToResourceMap, err := getNetworkToResourceMap(t.virtClient, vmi)
 	if err != nil {
@@ -933,10 +1045,21 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 		compute.Env = append(compute.Env, k8sv1.EnvVar{Name: varName, Value: resourceName})
 	}
 
-	if _, ok := vmi.Labels[debugLogs]; ok {
+	virtLauncherLogVerbosity := t.clusterConfig.GetVirtLauncherVerbosity()
+
+	if verbosity, isSet := vmi.Labels[logVerbosity]; isSet || virtLauncherLogVerbosity != virtconfig.DefaultVirtLauncherLogVerbosity {
+		// Override the cluster wide verbosity level if a specific value has been provided for this VMI
+		verbosityStr := fmt.Sprint(virtLauncherLogVerbosity)
+		if isSet {
+			verbosityStr = verbosity
+		}
+		compute.Env = append(compute.Env, k8sv1.EnvVar{Name: ENV_VAR_VIRT_LAUNCHER_LOG_VERBOSITY, Value: verbosityStr})
+	}
+
+	if labelValue, ok := vmi.Labels[debugLogs]; (ok && strings.EqualFold(labelValue, "true")) || virtLauncherLogVerbosity > EXT_LOG_VERBOSITY_THRESHOLD {
 		compute.Env = append(compute.Env, k8sv1.EnvVar{Name: ENV_VAR_LIBVIRT_DEBUG_LOGS, Value: "1"})
 	}
-	if _, ok := vmi.Labels[virtiofsDebugLogs]; ok {
+	if labelValue, ok := vmi.Labels[virtiofsDebugLogs]; (ok && strings.EqualFold(labelValue, "true")) || virtLauncherLogVerbosity > EXT_LOG_VERBOSITY_THRESHOLD {
 		compute.Env = append(compute.Env, k8sv1.EnvVar{Name: ENV_VAR_VIRTIOFSD_DEBUG_LOGS, Value: "1"})
 	}
 
@@ -1004,7 +1127,6 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 			nodeSelector[cpuFeatureLable] = "true"
 		}
 	}
-
 	if t.clusterConfig.HypervStrictCheckEnabled() {
 		hvNodeSelectors := getHypervNodeSelectors(vmi)
 		for k, v := range hvNodeSelectors {
@@ -1043,7 +1165,7 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 			Args:            requestedHookSidecar.Args,
 			Resources:       resources,
 			VolumeMounts: []k8sv1.VolumeMount{
-				k8sv1.VolumeMount{
+				{
 					Name:      "hook-sidecar-sockets",
 					MountPath: hooks.HookSocketsSharedDirectory,
 				},
@@ -1054,39 +1176,13 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 
 	hostName := dns.SanitizeHostname(vmi)
 
-	annotationsList := map[string]string{
-		v1.DomainAnnotation: domain,
-	}
-	if tempPod {
-		// mark pod as temp - only used for provisioning
-		annotationsList[v1.EphemeralProvisioningObject] = "true"
-	}
-	for k, v := range vmi.Annotations {
-		// filtering so users will not see this on pod and in confusion
-		if strings.HasPrefix(k, "kubectl.kubernetes.io") ||
-			strings.HasPrefix(k, "kubevirt.io/storage-observed-api-version") ||
-			strings.HasPrefix(k, "kubevirt.io/latest-observed-api-version") {
-			continue
-		}
-		annotationsList[k] = v
-	}
-
-	cniAnnotations, err := getCniAnnotations(vmi)
+	podAnnotations, err := generatePodAnnotations(vmi)
 	if err != nil {
 		return nil, err
 	}
-	for k, v := range cniAnnotations {
-		annotationsList[k] = v
-	}
-
-	for _, network := range vmi.Spec.Networks {
-		if network.Multus != nil && network.Multus.Default {
-			annotationsList[MULTUS_DEFAULT_NETWORK_CNI_ANNOTATION] = network.Multus.NetworkName
-		}
-	}
-
-	if HaveMasqueradeInterface(vmi.Spec.Domain.Devices.Interfaces) {
-		annotationsList[ISTIO_KUBEVIRT_ANNOTATION] = "k6t-eth0"
+	if tempPod {
+		// mark pod as temp - only used for provisioning
+		podAnnotations[v1.EphemeralProvisioningObject] = "true"
 	}
 
 	var initContainers []k8sv1.Container
@@ -1094,7 +1190,7 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 	if HaveContainerDiskVolume(vmi.Spec.Volumes) {
 
 		initContainerVolumeMounts := []k8sv1.VolumeMount{
-			k8sv1.VolumeMount{
+			{
 				Name:      "virt-bin-share-dir",
 				MountPath: "/init/usr/bin",
 			},
@@ -1140,7 +1236,7 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 		ObjectMeta: metav1.ObjectMeta{
 			GenerateName: "virt-launcher-" + domain + "-",
 			Labels:       podLabels,
-			Annotations:  annotationsList,
+			Annotations:  podAnnotations,
 			OwnerReferences: []metav1.OwnerReference{
 				*metav1.NewControllerRef(vmi, v1.VirtualMachineInstanceGroupVersionKind),
 			},
@@ -1150,7 +1246,6 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 			Subdomain: vmi.Spec.Subdomain,
 			SecurityContext: &k8sv1.PodSecurityContext{
 				RunAsUser: &userId,
-				FSGroup:   &t.launcherSubGid,
 			},
 			TerminationGracePeriodSeconds: &gracePeriodKillAfter,
 			RestartPolicy:                 k8sv1.RestartPolicyNever,
@@ -1167,24 +1262,7 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 	// If an SELinux type was specified, use that--otherwise don't set an SELinux type
 	selinuxType := t.clusterConfig.GetSELinuxLauncherType()
 	if selinuxType != "" {
-		pod.Spec.SecurityContext.SELinuxOptions = &k8sv1.SELinuxOptions{Type: selinuxType}
-		// By setting an SELinux option on the virt-launcher pod, we trigger this:
-		// https://github.com/kubernetes/kubernetes/issues/90759
-		// Since the compute container needs to be able to communicate with the rest of the pod,
-		//   we loop over all the containers and remove their SELinux categories.
-		for i := range pod.Spec.Containers {
-			container := &pod.Spec.Containers[i]
-			if container.Name != "compute" {
-				if container.SecurityContext == nil {
-					container.SecurityContext = &k8sv1.SecurityContext{}
-				}
-				if container.SecurityContext.SELinuxOptions == nil {
-					container.SecurityContext.SELinuxOptions = &k8sv1.SELinuxOptions{}
-				}
-				container.SecurityContext.SELinuxOptions.Type = selinuxType
-				container.SecurityContext.SELinuxOptions.Level = "s0"
-			}
-		}
+		alignPodMultiCategorySecurity(&pod, selinuxType)
 	}
 
 	if vmi.Spec.PriorityClassName != "" {
@@ -1218,8 +1296,24 @@ func (t *templateService) renderLaunchManifest(vmi *v1.VirtualMachineInstance, t
 	return &pod, nil
 }
 
-func (t *templateService) RenderHotplugAttachmentPodTemplate(volume *v1.Volume, ownerPod *k8sv1.Pod, vmi *v1.VirtualMachineInstance, pvcName string, isBlock bool) (*k8sv1.Pod, error) {
+func (t *templateService) RenderHotplugAttachmentPodTemplate(volume *v1.Volume, ownerPod *k8sv1.Pod, _ *v1.VirtualMachineInstance, pvcName string, isBlock bool, tempPod bool) (*k8sv1.Pod, error) {
 	zero := int64(0)
+	sharedMount := k8sv1.MountPropagationHostToContainer
+	var command []string
+	if tempPod {
+		command = []string{"/bin/bash",
+			"-c",
+			"exit", "0"}
+	} else {
+		command = []string{"/bin/sh", "-c", "/usr/bin/container-disk --copy-path /path/hp"}
+	}
+
+	annotationsList := make(map[string]string)
+	if tempPod {
+		// mark pod as temp - only used for provisioning
+		annotationsList[v1.EphemeralProvisioningObject] = "true"
+	}
+
 	pod := &k8sv1.Pod{
 		ObjectMeta: metav1.ObjectMeta{
 			GenerateName: "hp-volume-",
@@ -1233,13 +1327,14 @@ func (t *templateService) RenderHotplugAttachmentPodTemplate(volume *v1.Volume,
 			Labels: map[string]string{
 				v1.AppLabel: "hotplug-disk",
 			},
+			Annotations: annotationsList,
 		},
 		Spec: k8sv1.PodSpec{
 			Containers: []k8sv1.Container{
 				{
 					Name:    "hotplug-disk",
 					Image:   t.launcherImage,
-					Command: []string{"/bin/sh", "-c", "tail -f /dev/null"},
+					Command: command,
 					Resources: k8sv1.ResourceRequirements{ //Took the request and limits from containerDisk init container.
 						Limits: map[k8sv1.ResourceName]resource.Quantity{
 							k8sv1.ResourceCPU:    resource.MustParse("100m"),
@@ -1256,6 +1351,13 @@ func (t *templateService) RenderHotplugAttachmentPodTemplate(volume *v1.Volume,
 							Type:  t.clusterConfig.GetSELinuxLauncherType(),
 						},
 					},
+					VolumeMounts: []k8sv1.VolumeMount{
+						{
+							Name:             "hotplug-disks",
+							MountPath:        "/path",
+							MountPropagation: &sharedMount,
+						},
+					},
 				},
 			},
 			Affinity: &k8sv1.Affinity{
@@ -1303,31 +1405,46 @@ func (t *templateService) RenderHotplugAttachmentPodTemplate(volume *v1.Volume,
 			RunAsUser: &[]int64{0}[0],
 		}
 	} else {
-		pod.Spec.Containers[0].VolumeMounts = []k8sv1.VolumeMount{
-			{
-				Name:      volume.Name,
-				MountPath: "/pvc",
-			},
-		}
+		pod.Spec.Containers[0].VolumeMounts = append(pod.Spec.Containers[0].VolumeMounts, k8sv1.VolumeMount{
+			Name:      volume.Name,
+			MountPath: "/pvc",
+		})
 	}
 	return pod, nil
 }
 
-func getRequiredCapabilities(vmi *v1.VirtualMachineInstance) []k8sv1.Capability {
-	res := []k8sv1.Capability{}
-	if (len(vmi.Spec.Domain.Devices.Interfaces) > 0) ||
-		(vmi.Spec.Domain.Devices.AutoattachPodInterface == nil) ||
-		(*vmi.Spec.Domain.Devices.AutoattachPodInterface == true) {
-		res = append(res, CAP_NET_ADMIN)
+func getVirtiofsCapabilities() []k8sv1.Capability {
+	return []k8sv1.Capability{
+		"CHOWN",
+		"DAC_OVERRIDE",
+		"FOWNER",
+		"FSETID",
+		"SETGID",
+		"SETUID",
+		"MKNOD",
+		"SETFCAP",
 	}
+}
+
+func getRequiredCapabilities(vmi *v1.VirtualMachineInstance, config *virtconfig.ClusterConfig) []k8sv1.Capability {
+	capabilities := []k8sv1.Capability{}
+
 	// add a CAP_SYS_NICE capability to allow setting cpu affinity
-	res = append(res, CAP_SYS_NICE)
+	capabilities = append(capabilities, CAP_SYS_NICE)
 
 	// add CAP_SYS_ADMIN capability to allow virtiofs
 	if util.IsVMIVirtiofsEnabled(vmi) {
-		res = append(res, CAP_SYS_ADMIN)
+		capabilities = append(capabilities, CAP_SYS_ADMIN)
+		capabilities = append(capabilities, getVirtiofsCapabilities()...)
 	}
-	return res
+
+	// add SYS_RESOURCE capability to enable Live Migration for VM with SRIOV interfaces
+	// until https://bugzilla.redhat.com/show_bug.cgi?id=1916346 is resolved.
+	if config.SRIOVLiveMigrationEnabled() && util.IsSRIOVVmi(vmi) {
+		capabilities = append(capabilities, CAP_SYS_RESOURCE)
+	}
+
+	return capabilities
 }
 
 func getRequiredResources(vmi *v1.VirtualMachineInstance, useEmulation bool) k8sv1.ResourceList {
@@ -1367,7 +1484,7 @@ func appendUniqueImagePullSecret(secrets []k8sv1.LocalObjectReference, newsecret
 //
 // Note: This is the best estimation we were able to come up with
 //       and is still not 100% accurate
-func getMemoryOverhead(vmi *v1.VirtualMachineInstance) *resource.Quantity {
+func getMemoryOverhead(vmi *v1.VirtualMachineInstance, cpuArch string) *resource.Quantity {
 	domain := vmi.Spec.Domain
 	vmiMemoryReq := domain.Resources.Requests.Memory()
 
@@ -1416,6 +1533,13 @@ func getMemoryOverhead(vmi *v1.VirtualMachineInstance) *resource.Quantity {
 		overhead.Add(resource.MustParse("16Mi"))
 	}
 
+	// When use uefi boot on aarch64 with edk2 package, qemu will create 2 pflash(64Mi each, 128Mi in total)
+	// it should be considered for memory overhead
+	// Additional information can be found here: https://github.com/qemu/qemu/blob/master/hw/arm/virt.c#L120
+	if cpuArch == "arm64" {
+		overhead.Add(resource.MustParse("128Mi"))
+	}
+
 	// Additional overhead of 1G for VFIO devices. VFIO requires all guest RAM to be locked
 	// in addition to MMIO memory space to allow DMA. 1G is often the size of reserved MMIO space on x86 systems.
 	// Additial information can be found here: https://www.redhat.com/archives/libvir-list/2015-November/msg00329.html
@@ -1491,7 +1615,7 @@ func getNetworkToResourceMap(virtClient kubecli.KubevirtClient, vmi *v1.VirtualM
 	for _, network := range vmi.Spec.Networks {
 		if network.Multus != nil {
 			namespace, networkName := getNamespaceAndNetworkName(vmi, network.Multus.NetworkName)
-			crd, err := virtClient.NetworkClient().K8sCniCncfIoV1().NetworkAttachmentDefinitions(namespace).Get(networkName, metav1.GetOptions{})
+			crd, err := virtClient.NetworkClient().K8sCniCncfIoV1().NetworkAttachmentDefinitions(namespace).Get(context.Background(), networkName, metav1.GetOptions{})
 			if err != nil {
 				return map[string]string{}, fmt.Errorf("Failed to locate network attachment definition %s/%s", namespace, networkName)
 			}
@@ -1501,55 +1625,6 @@ func getNetworkToResourceMap(virtClient kubecli.KubevirtClient, vmi *v1.VirtualM
 	return
 }
 
-func getIfaceByName(vmi *v1.VirtualMachineInstance, name string) *v1.Interface {
-	for _, iface := range vmi.Spec.Domain.Devices.Interfaces {
-		if iface.Name == name {
-			return &iface
-		}
-	}
-	return nil
-}
-
-func getCniAnnotations(vmi *v1.VirtualMachineInstance) (cniAnnotations map[string]string, err error) {
-	ifaceListMap := make([]map[string]string, 0)
-	cniAnnotations = make(map[string]string, 0)
-
-	next_idx := 0
-	for _, network := range vmi.Spec.Networks {
-		// Set the type for the first network. All other networks must have same type.
-		if network.Multus != nil {
-			if network.Multus.Default {
-				continue
-			}
-			namespace, networkName := getNamespaceAndNetworkName(vmi, network.Multus.NetworkName)
-			ifaceMap := map[string]string{
-				"name":      networkName,
-				"namespace": namespace,
-				"interface": fmt.Sprintf("net%d", next_idx+1),
-			}
-			iface := getIfaceByName(vmi, network.Name)
-			if iface != nil && iface.MacAddress != "" {
-				// De-facto Standard doesn't define exact string format for
-				// MAC addresses pasted down to CNI.  Here we just pass through
-				// whatever the value our API layer accepted as legit.
-				// Note: while standard allows for 20-byte InfiniBand addresses,
-				// we forbid them in API.
-				ifaceMap["mac"] = iface.MacAddress
-			}
-			next_idx = next_idx + 1
-			ifaceListMap = append(ifaceListMap, ifaceMap)
-		}
-	}
-	if len(ifaceListMap) > 0 {
-		ifaceJsonString, err := json.Marshal(ifaceListMap)
-		if err != nil {
-			return map[string]string{}, fmt.Errorf("Failed to create JSON list from CNI interface map %s", ifaceListMap)
-		}
-		cniAnnotations[MultusNetworksAnnotation] = fmt.Sprintf("%s", ifaceJsonString)
-	}
-	return
-}
-
 func NewTemplateService(launcherImage string,
 	virtShareDir string,
 	virtLibDir string,
@@ -1560,7 +1635,8 @@ func NewTemplateService(launcherImage string,
 	persistentVolumeClaimCache cache.Store,
 	virtClient kubecli.KubevirtClient,
 	clusterConfig *virtconfig.ClusterConfig,
-	launcherSubGid int64) TemplateService {
+	launcherSubGid int64,
+	cpuArch string) TemplateService {
 
 	precond.MustNotBeEmpty(launcherImage)
 	svc := templateService{
@@ -1575,6 +1651,7 @@ func NewTemplateService(launcherImage string,
 		virtClient:                 virtClient,
 		clusterConfig:              clusterConfig,
 		launcherSubGid:             launcherSubGid,
+		cpuArch:                    cpuArch,
 	}
 	return &svc
 }
@@ -1595,3 +1672,74 @@ func copyProbe(probe *v1.Probe) *k8sv1.Probe {
 		},
 	}
 }
+
+func alignPodMultiCategorySecurity(pod *k8sv1.Pod, selinuxType string) {
+	pod.Spec.SecurityContext.SELinuxOptions = &k8sv1.SELinuxOptions{Type: selinuxType}
+	// more info on https://github.com/kubernetes/kubernetes/issues/90759
+	// Since the compute container needs to be able to communicate with the
+	// rest of the pod, we loop over all the containers and remove their SELinux
+	// categories.
+	for i := range pod.Spec.Containers {
+		container := &pod.Spec.Containers[i]
+		if container.Name != "compute" {
+			container.SecurityContext = generateContainerSecurityContext(selinuxType)
+		}
+	}
+}
+
+func generateContainerSecurityContext(selinuxType string) *k8sv1.SecurityContext {
+	return &k8sv1.SecurityContext{
+		SELinuxOptions: &k8sv1.SELinuxOptions{
+			Type:  selinuxType,
+			Level: "s0",
+		},
+	}
+}
+
+func generatePodAnnotations(vmi *v1.VirtualMachineInstance) (map[string]string, error) {
+	annotationsSet := map[string]string{
+		v1.DomainAnnotation: vmi.GetObjectMeta().GetName(),
+	}
+	for k, v := range filterVMIAnnotationsForPod(vmi.Annotations) {
+		annotationsSet[k] = v
+	}
+
+	multusAnnotation, err := generateMultusCNIAnnotation(vmi)
+	if err != nil {
+		return nil, err
+	}
+	if multusAnnotation != "" {
+		annotationsSet[MultusNetworksAnnotation] = multusAnnotation
+	}
+
+	if multusDefaultNetwork := lookupMultusDefaultNetworkName(vmi.Spec.Networks); multusDefaultNetwork != "" {
+		annotationsSet[MULTUS_DEFAULT_NETWORK_CNI_ANNOTATION] = multusDefaultNetwork
+	}
+
+	if HaveMasqueradeInterface(vmi.Spec.Domain.Devices.Interfaces) {
+		annotationsSet[ISTIO_KUBEVIRT_ANNOTATION] = "k6t-eth0"
+	}
+	return annotationsSet, nil
+}
+
+func lookupMultusDefaultNetworkName(networks []v1.Network) string {
+	for _, network := range networks {
+		if network.Multus != nil && network.Multus.Default {
+			return network.Multus.NetworkName
+		}
+	}
+	return ""
+}
+
+func filterVMIAnnotationsForPod(vmiAnnotations map[string]string) map[string]string {
+	annotationsList := map[string]string{}
+	for k, v := range vmiAnnotations {
+		if strings.HasPrefix(k, "kubectl.kubernetes.io") ||
+			strings.HasPrefix(k, "kubevirt.io/storage-observed-api-version") ||
+			strings.HasPrefix(k, "kubevirt.io/latest-observed-api-version") {
+			continue
+		}
+		annotationsList[k] = v
+	}
+	return annotationsList
+}
